/*
 * File: terminal.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.  File altered by C. Smith.
 * FN key work by Hellmuth Michaelis.
 *
 * MACOS port by Aria Stewart <aredridel@nbtsc.org>
 */

/**

Unix port file:

low-level character-handling fns (tt*)
terminal screen fns (sc*)

**/

#if MACOS
#define POSIX
#endif                          /* MACOS */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<curses.h>
#include	<term.h>
#include	"rho.h"
#include	"terminal.h"
#undef CTRL
#include	<sys/types.h>
#include	<sys/ioctl.h>
#include	<signal.h>
extern void check_resize(int sig);
#if MACOS
#include	<termios.h>
#else
#include	<termio.h>
#endif                          /* MACOS */

#undef CTRL                     /* Restore CTRL after sys/ioctl.h */
#define CTRL 0x0100

/** Parameters **/
#define NKEYENT		300     /* Number of keymap entries     */
#define NINCHAR		64      /* Input buffer size            */
#define NOUTCHAR	256     /* Output buffer size           */
#define NCAPBUF		1024    /* Termcap storage size         */
#define MARGIN		8       /* Margin size                  */
#define SCRSIZ		64      /* Scroll for margin            */
#define NPAUSE		10      /* # times thru update to pause */

/** constants **/

#define TIMEOUT		255     /* No character available       */

/** Type definitions **/
struct keyent {                 /* Key mapping entry            */
    struct keyent *samlvl;      /* Character on same level      */
    struct keyent *nxtlvl;      /* Character on next level      */
    unsigned char ch;           /* Character                    */
    int code;                   /* Resulting keycode            */
};
struct capbind {                /* Capability binding entry     */
    char *name;                 /* Termcap name                 */
    char *store;                /* Storage variable             */
};
struct keybind {                /* Keybinding entry             */
    char *name;                 /* Termcap name                 */
    int value;                  /* Binding value                */
};
char *reset = (char *) NULL;    /* reset string kjc */

/** Local variables **/
#if MACOS
static struct termios curterm;  /* Current modes                */
static struct termios oldterm;  /* Original modes               */
#else
static struct termio curterm;   /* Current modes                */
static struct termio oldterm;   /* Original modes               */
#endif                          /* MACOS */


static char tcapbuf[NCAPBUF];   /* Termcap character storage    */
#define CAP_CL		0       /* Clear to end of page         */
#define CAP_CM		1       /* Cursor motion                */
#define CAP_CE		2       /* Clear to end of line         */
#define CAP_SE		3       /* Standout ends                */
#define CAP_SO		4       /* Standout (reverse video)     */
#define CAP_IS		5       /* Initialize screen            */
#define CAP_KS		6       /* Keypad mode starts           */
#define CAP_KE		7       /* Keypad mode ends             */
#define CAP_VB		8       /* Visible bell                 */

/*@ignore@*/
static struct capbind capbind[] = {     /* Capability binding list      */
    {"cl"},                     /* Clear to end of page         */
    {"cm"},                     /* Cursor motion                */
    {"ce"},                     /* Clear to end of line         */
    {"se"},                     /* Standout ends                */
    {"so"},                     /* Standout (reverse video)     */
    {"is"},                     /* Initialize screen            */
    {"ks"},                     /* Keypad mode starts           */
    {"ke"},                     /* Keypad mode ends             */
    {"vb"},                     /* Visible bell                 */

};

/*@end@*/

static struct keybind keybind[] = {     /* Keybinding list              */
    {"bt", SHFT | CTRL | 'i'},  /* Back-tab key                 */
    {"k1", SPEC | '1'},         /* F1 key                       */
    {"k2", SPEC | '2'},         /* F2 key                       */
    {"k3", SPEC | '3'},         /* F3 key                       */
    {"k4", SPEC | '4'},         /* F4 key                       */
    {"k5", SPEC | '5'},         /* F5 key                       */
    {"k6", SPEC | '6'},         /* F6 key                       */
    {"k7", SPEC | '7'},         /* F7 key                       */
    {"k8", SPEC | '8'},         /* F8 key                       */
    {"k9", SPEC | '9'},         /* F9 key                       */
    {"k0", SPEC | '0'},         /* F0 or F10 key                */
    {"k;", SPEC | '0'},         /* F0 or F10 key        (kjc)   */
    {"F1", SHFT | SPEC | '1'},  /* Shift-F1 or F11 key          */
    {"F2", SHFT | SPEC | '2'},  /* Shift-F2 or F12 key          */
    {"F3", SHFT | SPEC | '3'},  /* Shift-F3 or F13 key          */
    {"F4", SHFT | SPEC | '4'},  /* Shift-F4 or F14 key          */
    {"F5", SHFT | SPEC | '5'},  /* Shift-F5 or F15 key          */
    {"F6", SHFT | SPEC | '6'},  /* Shift-F6 or F16 key          */
    {"F7", SHFT | SPEC | '7'},  /* Shift-F7 or F17 key          */
    {"F8", SHFT | SPEC | '8'},  /* Shift-F8 or F18 key          */
    {"F9", SHFT | SPEC | '9'},  /* Shift-F9 or F19 key          */
    {"FA", SHFT | SPEC | '0'},  /* Shift-F0 or F20 key          */
    {"kA", CTRL | 'O'},         /* Insert line key              */
    {"kb", CTRL | 'H'},         /* Backspace key                */
    {"kC", CTRL | 'L'},         /* Clear screen key             */
    {"kD", SPEC | 'D'},         /* Delete character key         */
    {"kd", SPEC | 'N'},         /* Down arrow key               */
    {"kE", CTRL | 'K'},         /* Clear to end of line key     */
    {"kF", CTRL | 'V'},         /* Scroll forward key           */
    {"kH", SPEC | '>'},         /* Home down key                */
    {"@7", SPEC | '>'},         /* Home down key        (kjc)   */
    {"kh", SPEC | '<'},         /* Home key                     */
    {"kI", SPEC | 'C'},         /* Insert character key         */
    {"kL", CTRL | 'K'},         /* Delete line key              */
    {"kl", SPEC | 'B'},         /* Left arrow key               */
    {"kN", SPEC | 'V'},         /* Next page key                */
    {"kP", SPEC | 'Z'},         /* Previous page key            */
    {"kR", CTRL | 'Z'},         /* Scroll backward key          */
    {"kr", SPEC | 'F'},         /* Right arrow key              */
    {"ku", SPEC | 'P'}          /* Up arrow key                 */
};

static int inbuf[NINCHAR];      /* Input buffer                 */
static int *inbufh = inbuf;     /* Head of input buffer         */
static int *inbuft = inbuf;     /* Tail of input buffer         */
static unsigned char outbuf[NOUTCHAR];  /* Output buffer                */
static unsigned char *outbuft = outbuf; /* Output buffer tail           */
static unsigned char keyseq[256];       /* Prefix escape sequence table */
static struct keyent keymap[NKEYENT];   /* Key map                      */
static struct keyent *nxtkey = keymap;  /* Next free key entry          */

/** Terminal definition block **/
int scopen(), scclose(), ttgetc(), ttputc(), ttflush();
int scmove(), sceeol(), sceeop(), scbeep(), screv();
int sckopen(), sckclose(), scnothing();


TERM term = {
    120,                        /* Maximum number of rows       */
    0,                          /* Current number of rows       */
    256,                        /* Maximum number of columns    */
    0,                          /* Current number of columns    */
    0, 0,                       /* upper left corner default screen */
    MARGIN,                     /* Margin for extending lines   */
    SCRSIZ,                     /* Scroll size for extending    */
    NPAUSE,                     /* # times thru update to pause */
    scopen,                     /* Open terminal routine        */
    scclose,                    /* Close terminal routine       */
    sckopen,                    /* Open keyboard routine        */
    sckclose,                   /* Close keyboard routine       */
    ttgetc,                     /* Get character routine        */
    ttputc,                     /* Put character routine        */
    ttflush,                    /* Flush output routine         */
    scmove,                     /* Move cursor routine          */
    sceeol,                     /* Erase to end of line routine */
    sceeop,                     /* Erase to end of page routine */
    sceeop,                     /* Clear the desktop            */
    scbeep,                     /* Beep! routine                */
    screv,                      /* Set reverse video routine    */
    scnothing,                  /* Set resolution routine       */

};



/*++

Open terminal device

++*/
int ttopen()
{

    /* Get modes */
#if MACOS
    if (tcgetattr(0, &oldterm)) {
#else
    if (ioctl(0, TCGETA, &oldterm)) {
#endif                          /* MACOS */
        perror("Cannot TCGETA");
        return (-1);
    }

    /* Save to original mode variable */
    curterm = oldterm;

    /* Set new modes */
    curterm.c_iflag &= ~(INLCR | ICRNL | IGNCR);
    curterm.c_lflag &= ~(ICANON | ISIG | ECHO);
    curterm.c_cc[VMIN] = 1;
    curterm.c_cc[VTIME] = 0;

    /* Set tty mode */
#if MACOS
    if (tcsetattr(0, TCSANOW, &curterm)) {
#else
    if (ioctl(0, TCSETA, &curterm)) {
#endif                          /* MACOS */
        perror("Cannot TCSETA");
        return (-1);
    }

    /* Success */
    return (0);
}


extern ssize_t write(int, const void *, size_t);
/*++

Close terminal device

++*/
int ttclose()
{
    /* Restore original terminal modes */
    if (reset != (char *) NULL)
        (void) write(1, reset, strlen(reset));

#if MACOS
    if (tcsetattr(0, TCSANOW, &oldterm))
#else
    if (ioctl(0, TCSETA, &oldterm))
#endif                          /* MACOS */
        return (-1);

    /* Success */
    return (0);
}

/*++

Flush output buffer to display

++*/
int ttflush()
{
    int len;

    /* Compute length of write */
    len = outbuft - outbuf;
    if (len == 0)
        return (0);

    /* Reset buffer position */
    outbuft = outbuf;

    /* Perform write to screen */
    return ((int) write(1, outbuf, (size_t) len) != len);
}

/*++

Put character onto display

++*/
int ttputc(ch)
char ch;                        /* Character to display         */
{
    /* Check for buffer full */
    if (outbuft == &outbuf[sizeof(outbuf)])
        ttflush();

    /* Add to buffer */
    *outbuft++ = (unsigned char) ch;

    return (0);
}

/*++

Add character sequence to keycode entry.
Helper for scopen

++*/
static void addkey(seq, fn)
char *seq;                      /* Character sequence           */
int fn;                         /* Resulting keycode            */
{
    int first;
    struct keyent *cur, *nxtcur;

    /* Skip on null sequences */
    if (!seq)
        return;

    /* Skip single character sequences */
    if (strlen(seq) < 2)
        return;

    /* If no keys defined, go directly to insert mode */
    first = 1;
    if (nxtkey != keymap) {

        /* Start at top of key map */
        cur = keymap;

        /* Loop until matches exhast */
        while (*seq) {

            /* Do we match current character */
            if (*seq == (char) cur->ch) {

                /* Advance to next level */
                seq++;
                cur = cur->nxtlvl;
                first = 0;
            } else {

                /* Try next character on same level */
                nxtcur = cur->samlvl;

                /* Stop if no more */
                if (nxtcur)
                    cur = nxtcur;
                else
                    break;
            }
        }
    }

    /* Check for room in keymap */
    if (strlen(seq) > NKEYENT - (nxtkey - keymap))
        return;

    /* If first character in sequence is inserted, add to prefix table */
    if (first)
        keyseq[(int) *seq] = 1;

    /* If characters are left over, insert them into list */
    for (first = 1; *seq; first = 0) {

        /* Make new entry */
        nxtkey->ch = (unsigned) *seq++;
        nxtkey->code = fn;

        /* If root, nothing to do */
        if (nxtkey != keymap) {

            /* Set first to samlvl, others to nxtlvl */
            if (first)
                cur->samlvl = nxtkey;
            else
                cur->nxtlvl = nxtkey;
        }

        /* Advance to next key */
        cur = nxtkey++;
    }
}


extern ssize_t read(int, void *, size_t);
/*++

Grab input characters, with wait

++*/
static unsigned char grabwait()
{

    unsigned char ch;

    /* Change mode, if necessary */
    if (curterm.c_cc[VTIME]) {
        curterm.c_cc[VMIN] = 1;
        curterm.c_cc[VTIME] = 0;
#if MACOS
        tcsetattr(0, TCSANOW, &curterm);
#else
        ioctl(0, TCSETA, &curterm);
#endif                          /* MACOS */
    }

    /* Perform read */
    if (read(0, &ch, 1) != 1) {
        puts("** Horrible read error occured **");
        exit(1);
    }

    /* Return new character */
    return (ch);
}

/*++

Grab input characters, short wait

++*/
static unsigned char grabnowait()
{
    int count;
    unsigned char ch;

    /* Change mode, if necessary */
    if (curterm.c_cc[VTIME] == 0) {
        curterm.c_cc[VMIN] = 0;
        curterm.c_cc[VTIME] = 5;
#if MACOS
        tcsetattr(0, TCSANOW, &curterm);
#else
        ioctl(0, TCSETA, &curterm);
#endif                          /* MACOS */
    }

    /* Perform read */
    count = (int) read(0, &ch, 1);
    if (count < 0) {
        puts("** Horrible read error occured **");
        exit(1);
    }
    if (count == 0)
        return (TIMEOUT);

    /* Return new character */
    return (ch);
}

/*++

Queue input character

++*/
static void qin(ch)
int ch;                         /* Character to add             */
{
    /* Check for overflow */
    if (inbuft == &inbuf[sizeof(inbuf)]) {

        /* Annoy user */
        scbeep();
        return;
    }

    /* Add character */
    *inbuft++ = ch;
}

/*++

Cook input characters

++*/
static void cook()
{
    unsigned char ch;
    struct keyent *cur;

    /* Get first character untimed */
    ch = grabwait();
    qin(ch);

    /* Skip if the key isn't a special leading escape sequence */
    if (keyseq[ch] == 0) {
        /* if null, change it to (0/1/32) */
        if (ch == 0) {
            qin(CTRL >> 8);     /* control */
            qin(32);            /* space */
        }
        return;
    }

    /* Start at root of keymap */
    cur = keymap;

    /* Loop until keymap exhasts */
    while (cur) {

        /* Did we find a matching character */
        if (cur->ch == ch) {

            /* Is this the end */
            if (cur->nxtlvl == NULL) {

                /* Replace all character with new sequence */
                inbuft = inbuf;
                qin(cur->code);
                return;
            } else {
                /* Advance to next level */
                cur = cur->nxtlvl;

                /* Get next character, timed */
                ch = grabnowait();
                if (ch == TIMEOUT)
                    return;

                /* Queue character */
                qin(ch);
            }
        } else
            /* Try next character on same level */
            cur = cur->samlvl;
    }
}

/*++

Return cooked characters

++*/
int ttgetc()
{
    int ch;

    /* Loop until character is in input buffer */
    while (inbufh == inbuft)
        cook();

    /* Get input from buffer, now that it is available */
    ch = *inbufh++;

    /* reset us to the beginning of the buffer if there are no more
       pending characters */
    if (inbufh == inbuft)
        inbufh = inbuft = inbuf;

    /* Return next character */
    return (ch);
}


/*++

In the type-ahead detection code, the individual UNIX
system either has a FIONREAD or a FIORDCHK ioctl call.
Hopefully, your system uses one of them and this be detected
correctly without any intervention.

++*/
int type_ahead()
{
    int count;

    /* See if internal buffer is non-empty */
    if (inbufh != inbuft)
        return (1);

    /* Now check with system */
#ifdef FIONREAD                 /* Watch out!  This could bite you! */
    /* Get number of pending characters */
    if (ioctl(0, FIONREAD, &count))
        return (0);
    return (count);
#else                           /* not FIONREAD */

    /* Ask hardware for count */
    count = ioctl(0, FIORDCHK, 0);
    if (count < 0)
        return (0);
    return (count);

#endif                          /* FIONREAD */
}


/*++

Put out sequence, with padding

++*/
static void putpad(seq)
char *seq;                      /* Character sequence */
{
    /* Check for null */
    if (!seq)
        return;

    /* Call on termcap to send sequence */
    tputs(seq, 1, ttputc);
}


/*++

Initialize screen package

++*/
int scopen()
{
    char *cp, termenv[256], tcbuf[1024];
    size_t limit = 256;
    int status;
    struct capbind *cb;
    struct keybind *kp;
    char err_str[NSTRING];


#define TGETSTR(a,b)	tgetstr((a), (b))

    char *UP;

    /* Get terminal type */
    strcpy(termenv, getenv("TERM"));
    if (!*termenv) {
        puts(TEXT182);          /*"Environment variable TERM not defined" */
        exit(1);
    }

    /* Try to load termcap */
    status = tgetent(tcbuf, termenv);
    if (status == -1) {
        puts("Cannot open termcap file");
        exit(1);
    }
    if (status == 0) {
        snprintf(err_str, limit, TEXT183, termenv);     /*"No entry for terminal type */
        puts(err_str);
        exit(1);
    }


    (void) signal(SIGWINCH, &check_resize);

    /* Get size from termcap */
    term.t_nrow = tgetnum("li") - 1;
    term.t_ncol = tgetnum("co");
    if (term.t_nrow < 3 || term.t_ncol < 3) {
        puts("Screen size is too small!");
        exit(1);
    }

    /* initialize max number of rows and cols        */
    term.t_mrow = 120;
    term.t_mcol = 256;

    /* Start grabbing termcap commands */
    cp = tcapbuf;

    /* Get the reset string */
    reset = TGETSTR("is", &cp);

    /* Get up line capability */
    UP = TGETSTR("up", &cp);

    /* Get other capabilities */
    cb = capbind;
    while (cb < &capbind[sizeof(capbind) / sizeof(*capbind)]) {
        cb->store = TGETSTR(cb->name, &cp);
        cb++;
    }

    /* Check for minimum */
    if (!capbind[CAP_CL].store && (!capbind[CAP_CM].store || !UP)) {
        puts("This terminal doesn't have enough power to run Rho!");
        exit(1);
    }

    /* Set reverse video and erase to end of line */
    if (capbind[CAP_SO].store && capbind[CAP_SE].store)
        revexist = TRUE;
    if (!capbind[CAP_CE].store)
        eolexist = FALSE;

    /* Get keybindings */
    kp = keybind;
    while (kp < &keybind[sizeof(keybind) / sizeof(*keybind)]) {
        addkey(TGETSTR(kp->name, &cp), kp->value);
        kp++;
    }

    /* Open terminal device */
    if (ttopen()) {
        puts("Cannot open terminal");
        exit(1);
    }

    /* Send out initialization sequences */
    putpad(capbind[CAP_IS].store);
    putpad(capbind[CAP_KS].store);
    sckopen();


    /* Success */
    return (0);
}

/*++

Close screen package

++*/
int scclose()
{
    /* Turn off keypad mode */
    putpad(capbind[CAP_KE].store);
    sckclose();

    /* Close terminal device */
    ttflush();
    ttclose();

    /* Success */
    return (0);
}


/*++

open keyboard -hm

++*/
int sckopen()
{
    putpad(capbind[CAP_KS].store);
    ttflush();
    return (0);
}


/*++

close keyboard -hm

++*/
int sckclose()
{
    putpad(capbind[CAP_KE].store);
    ttflush();
    return (0);
}


/*++

Move cursor

++*/
int scmove(row, col)
int row;                        /* Row number                   */
int col;                        /* Column number                */
{

    /* Call on termcap to create move sequence */
    putpad(tgoto(capbind[CAP_CM].store, col, row));

    /* Success */
    return (0);
}


/*++

Erase to end of line

++*/
int sceeol()
{
    /* Send erase sequence */
    putpad(capbind[CAP_CE].store);

    /* Success */
    return (0);
}

/*++

Clear screen

++*/
int sceeop()
{
    /* Send clear sequence */
    putpad(capbind[CAP_CL].store);

    /* Success */
    return (0);
}

/*++

Set reverse video state

++*/
int screv(state)
int state;                      /* New state                    */
{

    /* Set reverse video state */
    putpad(state ? capbind[CAP_SO].store : capbind[CAP_SE].store);

    /* Success */
    return (0);
}

/*++

Beep

++*/
int scbeep()
{
    /* Send out visible bell, if it exists */
    if (capbind[CAP_VB].store)
        putpad(capbind[CAP_VB].store);
    else
        /* The old standby method */
        ttputc('\7');

    /* Success */
    return (0);
}



/*++

Do nothing routine for the sc*

++*/
int scnothing()
{
    return (0);
}




/*++

Callout to system to perform command

++*/
int callout(cmd)
char *cmd;                      /* Command to execute           */
{
    int status;

    /* Close down */
    scmove(term.t_nrow, 0);
    ttflush();
    sckclose();
    ttclose();

    /* Do command */
    status = system(cmd) == 0;

    /* Restart system */
    sgarbf = TRUE;
    sckopen();
    if (ttopen()) {
        puts("** Error reopening terminal device **");
        exit(1);
    }

    /* Success */
    return (status);
}
